#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifdef JFB_ENABLE_SCROLLBACK
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>

#include <linux/kd.h>
#include <linux/keyboard.h>

#include "fbcommon.h"
#include "font.h"
#include "vterm.h"


/*
 * retrieve histlen size
 */

#define HISTLEN_DEFAULT 256

static int histlen = 0;

void sb_set_histlen(int len)
{
	if(0 < len && len < 10000)
		histlen = len;
	else
		histlen = HISTLEN_DEFAULT;
}

int sb_get_histlen(void)
{
	return histlen;
}


/*
 * keybord manipulation
 */

#define SHIFT_TABLE 1
#define PAGE_UP_KEYCODE 104
#define PAGE_DOWN_KEYCODE 109
#define PAGE_UP_ACTION 0x0118
#define PAGE_DOWN_ACTION 0x0119

u_short scroll_forward = 0;
u_short scroll_backward = 0;

int sb_set_kb(u_short table, u_short keycode, u_short value, u_short *oldvalue, u_short readonly)
{
	struct kbentry ke;
	ke.kb_table = table;
	ke.kb_index = keycode;

	if(ioctl(fileno(stdin), KDGKBENT, &ke))
		return -1;

	if(readonly == 0 && oldvalue)
		*oldvalue = ke.kb_value;

	ke.kb_value = value;

	if(ioctl(fileno(stdin), KDSKBENT, &ke))
		return -1;

	return 0;
}

void sb_restore_key_atexit(void)
{
	if(scroll_backward)
		sb_set_kb(SHIFT_TABLE, PAGE_UP_KEYCODE, scroll_backward, 0, 1);
	if(scroll_forward)
		sb_set_kb(SHIFT_TABLE, PAGE_DOWN_KEYCODE, scroll_forward, 0, 1);
}

void sb_restore_key(void)
{
	if(scroll_backward)
		sb_set_kb(SHIFT_TABLE, PAGE_UP_KEYCODE, scroll_backward, 0, 1);
	if(scroll_forward)
		sb_set_kb(SHIFT_TABLE, PAGE_DOWN_KEYCODE, scroll_forward, 0, 1);
}

int sb_disable_key(u_short readonly)
{
	if(sb_set_kb(SHIFT_TABLE, PAGE_UP_KEYCODE, PAGE_UP_ACTION, &scroll_backward, readonly))
		return -1;
	if(sb_set_kb(SHIFT_TABLE, PAGE_DOWN_KEYCODE, PAGE_DOWN_ACTION, &scroll_forward, readonly))
		return -1;
	atexit(sb_restore_key_atexit);
	return 0;
}


/*
 * scrollback memory management
 */

static inline u_int sb_coord_to_index_hist(TVterm *p, u_int x, u_int y)
{
	return (p->textHead + x + y * p->xcap4) % (p->xcap4 * p->histsize);
}

/* same as tvterm_coord_to_index */
static inline u_int sb_coord_to_index(TVterm *p, u_int x, u_int y)
{
	return (p->textHead + x + y * p->xcap4) % p->tsize;
}

void sb_move_to_hist(TVterm *p, int dst, int src, int n)
{
	memmove(p->htext+dst, p->text+src, n*sizeof(u_int));
	memmove(p->hattr+dst, p->attr+src, n);
	memmove(p->hflag+dst, p->flag+src, n);
}

void sb_move_from_hist(TVterm *p, int dst, int src, int n)
{
	memmove(p->text+dst, p->htext+src, n*sizeof(u_int));
	memmove(p->attr+dst, p->hattr+src, n);
	memmove(p->flag+dst, p->hflag+src, n);
}

/*
 * in-only circular queue
 */

static int queue_full = 0;
static int queue_ptr = 0;

int
sb_queue_get_size(TVterm *p)
{
        if(queue_full != 0)
                return p->histsize;
	return queue_ptr;
}

int
sb_queue_line_rel_to_abs(TVterm *p, int rel)
{
	if(rel > sb_queue_get_size(p) || rel < 0) {
		return -1;
	}
	return (queue_ptr - rel + p->histsize) % p->histsize;
}

void
sb_queue_insert_line(TVterm *p, int src)
{
	int dst;
	dst = sb_coord_to_index_hist(p, 0, queue_ptr);
	sb_move_to_hist(p, dst, src, p->xcap4);

	queue_ptr = ++queue_ptr % p->histsize;
	if(queue_ptr == (p->histsize - 1))
		queue_full = 1;
}

/*
 * scrollback refresh routine
 */

void
sb_clean_hist(TVterm *p, u_int from, u_int to)
{
	u_int x, y, i;
	for(y = from; y < to; y++) {
		for(x = 0; x < p->xcap; x++) {
			i = sb_coord_to_index_hist(p, x, y);
			p->hflag[i] = p->hflag[i] & ~CLEAN_S;
		}
	}
}

void
sb_clean(TVterm *p, u_int from, u_int to)
{
	u_int x, y, i;
	for(y = from; y < to; y++) {
		for(x = 0; x < p->xcap; x++) {
			i = sb_coord_to_index(p, x, y);
			p->flag[i] = p->flag[i] & ~CLEAN_S;
		}
	}
}

void
sb_draw_chars_hist(TVterm *p, u_int from, u_int to, u_int draw_offset)
{
	TFont *pf;
	u_int x, y, i;
	u_int w, gw, chlw, lang;
	u_char bc, fc, fg, ch, ch2;
	const u_char *glyph;
	for(y = from; y < to; y++) {
		for(x = 0; x < p->xcap; x++) {
			i = sb_coord_to_index_hist(p, x, y);
			w = 1;
			fg = p->hflag[i];
			if(fg & CLEAN_S) {
				continue;
			}
			fc = p->hattr[i] & 0xf;
			bc = p->hattr[i] >> 4;
			chlw = p->htext[i];
			ch2 = (chlw >> 8) & 0xff;
			ch = chlw & 0xff;
			lang = (chlw >> 24) & 0xff;
			p->hflag[i] |= CLEAN_S;
			if(fg & CODEIS_1) {
				pf = &(gFont[lang]);
				i++;
				p->hflag[i] |= CLEAN_S;
				w = 2;
			}
			else
				pf = &(gFont[lang]);
			glyph = pf->conv(pf, chlw, &gw);
			gFramebuffer.cap.fill(&gFramebuffer,
				gFontsWidth * x, gFontsHeight * ((y - from) + draw_offset),
				gFontsWidth * w, gFontsHeight, bc);
			if(chlw == 0)
				continue;
			gFramebuffer.cap.overlay(&gFramebuffer,
				gFontsWidth * x, gFontsHeight * ((y - from) + draw_offset),
				glyph, gw, pf->height, pf->bytew, fc);
		}
	}
}

void
sb_draw_chars(TVterm *p, u_int from, u_int to, u_int draw_offset)
{
	TFont *pf;
	u_int x, y, i;
	u_int w, gw, chlw, lang;
	u_char bc, fc, fg, ch, ch2;
	const u_char *glyph;
	for(y = from; y < to; y++) {
		for(x = 0; x < p->xcap; x++) {
			i = sb_coord_to_index(p, x, y);
			w = 1;
			fg = p->flag[i];
			if(fg & CLEAN_S) {
				continue;
			}
			fc = p->attr[i] & 0xf;
			bc = p->attr[i] >> 4;
			chlw = p->text[i];
			ch2 = (chlw >> 8) & 0xff;
			ch = chlw & 0xff;
			lang = (chlw >> 24) & 0xff;
			p->flag[i] |= CLEAN_S;
			if(fg & CODEIS_1) {
				pf = &(gFont[lang]);
				i++;
				p->flag[i] |= CLEAN_S;
				w = 2;
			}
			else
				pf = &(gFont[lang]);
			glyph = pf->conv(pf, chlw, &gw);
			gFramebuffer.cap.fill(&gFramebuffer,
				gFontsWidth * x, gFontsHeight * ((y - from) + draw_offset),
				gFontsWidth * w, gFontsHeight, bc);
			if(chlw == 0)
				continue;
			gFramebuffer.cap.overlay(&gFramebuffer,
				gFontsWidth * x, gFontsHeight * ((y - from) + draw_offset),
				glyph, gw, pf->height, pf->bytew, fc);
		}
	}
}

void sb_refresh(TVterm *p, int page)
{
	int queue_size;
	int page_top, page_bot;
	int page_top_abs, page_bot_abs;
	
	tvterm_hide_cursor(p);
	page_top = (p->ycap / 2) * page;
	queue_size = sb_queue_get_size(p);
	if(page_top > queue_size) page_top = queue_size;
	if(page_top < p->ycap) page_bot = 0;
	else page_bot = page_top - p->ycap;

	page_top_abs = sb_queue_line_rel_to_abs(p, page_top);
	page_bot_abs = sb_queue_line_rel_to_abs(p, page_bot);

	if(page_top_abs >= page_bot_abs) {
		sb_clean_hist(p, page_top_abs, queue_size+1);
		sb_clean_hist(p, 0, page_bot_abs+1);

		sb_draw_chars_hist(p, page_top_abs, queue_size, 0);
		sb_draw_chars_hist(p, 0, page_bot_abs+1, queue_size - page_top_abs);
	}
	else if(page_top_abs < page_bot_abs) { 
		sb_clean_hist(p, page_top_abs, page_bot_abs+1);
		sb_draw_chars_hist(p, page_top_abs, page_bot_abs+1, 0);
	}

	sb_clean(p, 0, (p->ycap > page_top) ? (p->ycap - page_top): 0);
	sb_draw_chars(p, 0, (p->ycap > page_top) ? (p->ycap - page_top) : 0, page_top);
	sb_clean(p, 0, p->ycap);
}

/*
 * scrollback actions
 */

static int pageno = 0;

void sb_init_pageno(void)
{
	pageno = 0;
}

void sb_scroll_up(TVterm *p)
{
	int max_scroll_pages;
	max_scroll_pages = (sb_queue_get_size(p) + (p->ycap / 2)) / (p->ycap / 2);
	if(max_scroll_pages < 0) max_scroll_pages = 0;
	if(max_scroll_pages > pageno) {
		sb_refresh(p, ++pageno);
	}
}

void sb_scroll_down(TVterm *p)
{
	if(pageno == 1) {
		pageno = 0;
		//tvterm_refresh(p);
		sb_clean(p, 0, p->ycap);
		sb_draw_chars(p, 0, p->ycap, 0);
	}
	else if(pageno > 1) {
		sb_refresh(p, --pageno);
	}
}

int sb_is_on(void)
{
	return (pageno == 0) ? 0 : 1;
}
#endif
